// ==============================================================
//
// f9_v12_stg1_expendable
// Control module for f9_v12_stg1_expendable vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>

// ==============================================================
// Some vessel parameters
// ==============================================================


const double DRYMASS = 22100;//19000;
const VECTOR3 INT_OFF = _V(0,0,0);


static const DWORD ntdvtx_gearup = 4;
static TOUCHDOWNVTX tdvtx_gearup[ntdvtx_gearup] = {
	{_V(0, -5,-20), 1e6, 1e5, 1, 1},
	{_V(-4.33, 2.5,-20), 1e6, 1e5, 1, 1},
	{_V( 4.33, 2.5,-20), 1e6, 1e5, 1, 1},
	{_V(0 ,0 , 20), 1e6, 1e5, 1}
};

void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0, -0.1, -0.2, -0.1, 0};


	aoa += PI;
	int idx = max (0, min (15, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = 0.0;
	*cd = 0.55*(0.11 + oapiGetInducedDrag (*cl, 0.16, 0.2));
	
	
}

// 2. horizontal lift component (vertical stabiliser and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0, -0.1, -0.2, -0.1, 0};

	beta += PI;
	int idx = max (0, min (15, (int)(beta*istep)));
	double d = beta*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = 0.0;
	*cd = 0.55*(0.11 + oapiGetInducedDrag (*cl, 0.16, 0.2));
	
}

// ==============================================================
// Shuttle-PB class interface
// ==============================================================

class f9_v12_stg1_expendable: public VESSEL3 {
public:
	f9_v12_stg1_expendable (OBJHANDLE hVessel, int flightmodel);
	~f9_v12_stg1_expendable ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPostCreation();
	void clbkPreStep(double SimT, double SimDT, double MJD);
	
private:


};

f9_v12_stg1_expendable::f9_v12_stg1_expendable (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

f9_v12_stg1_expendable::~f9_v12_stg1_expendable ()
{
	

}


// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void f9_v12_stg1_expendable::clbkSetClassCaps (FILEHANDLE cfg)
{
	
	// physical vessel parameters
	SetSize (20);
	SetEmptyMass (DRYMASS);
	SetPMI(_V(129,129,1.9));
	SetCrossSections (_V(159,159,16));
	SetCW(0.5,4,1,1);
	SetRotDrag (_V(0.25,0.25,0.1));
	SetTouchdownPoints ( tdvtx_gearup, ntdvtx_gearup);

	// camera parameters
	SetCameraOffset (_V(0,3,7.5));
	SetCameraRotationRange(PI,PI,(0.4*PI),(0.4*PI));

	SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/f9_v12_stg1_nolegs"),MESHVIS_ALWAYS);

	oapiLoadMesh("Falcon9_BLOCK4/f9_v12_interstg_nolegs");
	SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/f9_v12_interstg_nolegs", &INT_OFF),MESHVIS_ALWAYS);

	
	// propellant resources
	


	// ***************** thruster definitions *******************

	

	///////////////////////////////////////////////////////
	/////AirFoils

	CreateAirfoil (LIFT_VERTICAL, _V(0,0,0.1), VLiftCoeff, 25, 80, 0.16);
	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,0.1), HLiftCoeff, 25, 80, 0.16); //90
	////////

	
	

}

void f9_v12_stg1_expendable::clbkPostCreation ()
{

}


// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new f9_v12_stg1_expendable (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (f9_v12_stg1_expendable*)vessel;
}



void f9_v12_stg1_expendable::clbkPreStep(double SimT, double SimDT, double MJD)
{
	if(GetAltitude(ALTMODE_GROUND) < 500)oapiDeleteVessel(GetHandle());
}





