// ==============================================================
//
// og2
// Control module for og2 vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>

// ==============================================================
// Some vessel parameters
// ==============================================================



const double ISP = 2250;
const double RCSTH = 25;
const double FUELMASS = 12;
const double DRYMASS = 160;


const double MAST_SPEED = 0.1;
const double SOLP_SPEED = 0.05;
const double ST_SPEED = 0.1;

#define GREEN RGB(125, 255, 125)


static const DWORD ntdvtx_geardown = 3;
static TOUCHDOWNVTX tdvtx_geardown[ntdvtx_geardown] = {
	{_V( 0,-1, 5), 1e6, 1e6, 0.7, 0.7},
	{_V(-2,-1,-5), 1e6, 1e6, 0.7, 0.7},
	{_V( 2,-1,-5), 1e6, 1e6, 0.7, 0.7}
};

void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = CM[i] + (CM[i+1]-CM[i]) * f;  // aoa-dependent moment coefficient
	double saoa = sin(aoa);
	double pd = 0.03 + 0.4*saoa*saoa;  // profile drag
	*cd = pd + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	// profile drag + (lift-)induced drag + transonic/supersonic wave (compressibility) drag
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.03 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}
// ==============================================================
// Shuttle-PB class interface
// ==============================================================

class og2: public VESSEL3 {
public:
	og2 (OBJHANDLE hVessel, int flightmodel);
	~og2 ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);


private:

	THRUSTER_HANDLE th_rcs[4], th_group[4];
	PROPELLANT_HANDLE hProp;

	ANIMATIONCOMPONENT_HANDLE Section, Section1, Section2;

	void DefineAnimations (void);
	double step_time;
	int kr_status, solp_status, mast_status, st_status, asp_status;

	UINT anim_solp, anim_mast, anim_st;
	double solp_proc, mast_proc, st_proc;
	void DoSunPointing(void);
	

	MGROUP_TRANSFORM *Solp1rot, *Solp2rot, *Solp3rot, *STrot;
	MGROUP_TRANSFORM *AntTrans, *AntScale;

	
};

og2::og2 (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

og2::~og2 ()
{
	delete Solp1rot;
	delete Solp2rot;
	delete Solp3rot;
	delete STrot;
	delete AntTrans;
	delete AntScale;


}

void og2::DefineAnimations (void)
{
// 1. SOLP

	anim_st = CreateAnimation (0);
	static UINT STGrp[1] = {8};
	STrot = new MGROUP_ROTATE(0, STGrp, 0, _V(0, 0.525, 0.265), _V(1,0,0), (float)(-90*RAD));
	Section = AddAnimationComponent (anim_st, 0, 1, STrot);

	anim_solp = CreateAnimation (0);
	static UINT Solp1Grp[1] = {8};
	Solp1rot = new MGROUP_ROTATE(0, Solp1Grp, 1, _V(0, 0.525, 0.265), _V(1,0,0), (float)(-90*RAD));
	Section1 = AddAnimationComponent (anim_solp, 0, 1, Solp1rot, Section);


	static UINT Solp2Grp[1] = {7};
	Solp2rot = new MGROUP_ROTATE(0, Solp2Grp, 1, _V(0,-0.518, 0.284), _V(1,0,0), (float)(180*RAD));
	Section2 = AddAnimationComponent (anim_solp, 0, 1, Solp2rot, Section1);

	
	static UINT Solp3Grp[1] = {6};
	Solp3rot = new MGROUP_ROTATE(0, Solp3Grp, 1, _V(0, 0.516, 0.302), _V(1,0,0), (float)(-180*RAD));
	AddAnimationComponent (anim_solp, 0, 1, Solp3rot, Section2);

// Mast

	anim_mast = CreateAnimation (0);
	static UINT AntMastGrp[1] = {5};
	AntScale = new MGROUP_SCALE (0, AntMastGrp, 1,_V(0,0,0), _V(1, 0.07, 1));
	
	static UINT AppTipGrp[1] = {4};
	AntTrans = new MGROUP_TRANSLATE(0, AppTipGrp, 1, _V(0, 7.068, 0));
	
	AddAnimationComponent (anim_mast, 0, 1, AntScale);
	AddAnimationComponent (anim_mast, 0, 1, AntTrans);


}


// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void og2::clbkSetClassCaps (FILEHANDLE cfg)
{
	step_time = 0;
	kr_status = 0;

	solp_status = 0;
	mast_status = 2;
	st_status = 0;

	asp_status = 0;

	solp_proc = 0.0;
	mast_proc = 1.0;
	st_proc = 0.0;

	DefineAnimations ();
	// physical vessel parameters
	SetSize (2);
	SetEmptyMass (DRYMASS);
	SetPMI(_V(2,2,0.75));
	SetCrossSections (_V(2,2,2));
	SetCW(1,1,1,1);
	SetRotDrag (_V(2,2,1));
	//SetTouchdownPoints ( _V(0,-1, 5), _V(-2,-1,-5), _V( 2,-1,-5));
	SetTouchdownPoints(tdvtx_geardown,ntdvtx_geardown);

	// camera parameters
	SetCameraOffset (_V(0,1.5,-1));

	SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/og2"),MESHVIS_EXTERNAL);


	// propellant resources
	
	hProp = CreatePropellantResource (FUELMASS,FUELMASS);

	// ***************** thruster definitions *******************
	

	th_rcs[0] = CreateThruster (_V( 0.509, 0.485, 0.092), _V(-0.5,-0.5, 0.707), RCSTH, hProp, ISP);
	th_rcs[1] = CreateThruster (_V(-0.509, 0.485, 0.092), _V( 0.5,-0.5, 0.707), RCSTH, hProp, ISP);
	th_rcs[2] = CreateThruster (_V( 0.509,-0.485, 0.092), _V(-0.5, 0.5, 0.707), RCSTH, hProp, ISP);
	th_rcs[3] = CreateThruster (_V(-0.509,-0.485, 0.092), _V( 0.5, 0.5, 0.707), RCSTH, hProp, ISP);

	

	th_group[0] = th_rcs[2];
	th_group[1] = th_rcs[3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[1];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[2];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[1];
	th_group[2] = th_rcs[2];
	th_group[3] = th_rcs[3];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_FORWARD);


	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[1];
	th_group[2] = th_rcs[2];
	th_group[3] = th_rcs[3];
	CreateThrusterGroup (th_group, 4, THGROUP_MAIN);

	SURFHANDLE exhaust = oapiRegisterExhaustTexture ("exhaust_atrcs");


	double exlen = 0.5;
	double exwid = 0.05;

	AddExhaust(th_rcs[0],exlen,exwid,exhaust);
	AddExhaust(th_rcs[1],exlen,exwid,exhaust);
	AddExhaust(th_rcs[2],exlen,exwid,exhaust);
	AddExhaust(th_rcs[3],exlen,exwid,exhaust);


	///////////////////////////////////////////////////////
	/////AirFoils

	CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,1), VLiftCoeff, 0, 1, 2, 1);
	CreateAirfoil3 (LIFT_HORIZONTAL, _V(1,0,0), HLiftCoeff, 0, 1, 2, 1);
}



// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new og2 (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (og2*)vessel;
}


void og2::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState(scn);

	sprintf (cbuf, "%d %0.4f %d %0.4f", solp_status, solp_proc, st_status, st_proc);
	oapiWriteScenario_string (scn, "SOLP", cbuf);

	sprintf (cbuf, "%d %0.4f", mast_status, mast_proc);
	oapiWriteScenario_string (scn, "LDMA", cbuf);
	
	sprintf (cbuf, "%d %d", kr_status, asp_status);
	oapiWriteScenario_string (scn, "KROT", cbuf);


}
void og2::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "SOLP", 4)) {
			sscanf (line+4, "%d%lf%d%lf", &solp_status, &solp_proc, &st_status, &st_proc);
		} else if (!strnicmp (line, "LDMA", 4)) {
			sscanf (line+4, "%d%lf", &mast_status, &mast_proc);
		} else if (!strnicmp (line, "KROT", 4)) {
			sscanf (line+4, "%d%d",&kr_status, &asp_status);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }

	SetAnimation(anim_solp, solp_proc);
	SetAnimation(anim_st, st_proc);
	SetAnimation(anim_mast, mast_proc);



}



int og2::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {



if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {	 
	switch (key)
	{
		

	case OAPI_KEY_NUMPAD5: // killrot
		if (kr_status == 0)
		{
		asp_status = 0;
		kr_status = 1;
		}
		else if (kr_status == 1) kr_status = 0;
		return 1;

	} 
  }

  else if (KEYMOD_CONTROL (kstate)) 
  { 
	switch (key)
	{
		
	} 
  }

else { // unmodified keys
    switch (key) {

		case OAPI_KEY_G: // solps
		if (solp_status == 0) solp_status = 1;
		if (mast_status == 2) mast_status = 1;
		return 1;

		case OAPI_KEY_B: // auto sun pointing
		if (asp_status == 0)
			{		
			kr_status = 0;
			asp_status = 1;
			}
		else asp_status = 0;
		return 1;


	}
  }
  return 0;
}


int og2::clbkConsumeDirectKey(char *kstate)
{


VECTOR3 avel = _V(0,0,0);
GetAngularVel(avel);
double dt = oapiGetSimStep();
double drot = RAD*2*dt;

if(kr_status == 0)
{
if (KEYMOD_SHIFT (kstate))
	{
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD2))
		{
		if (avel.x < RAD*10)
			{
			avel.x = avel.x+drot;
			if (avel.x > RAD*10) avel.x = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD8))
		{
		if (avel.x > -RAD*10)
			{
			avel.x = avel.x-drot;
			if (avel.x < -RAD*10) avel.x = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD1))
		{
		if (avel.y < RAD*10)
			{
			avel.y = avel.y+drot;
			if (avel.y > RAD*10) avel.y = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD3))
		{
		if (avel.y > -RAD*10)
			{
			avel.y = avel.y-drot;
			if (avel.y < -RAD*10) avel.y = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD6))
		{
		if (avel.z < RAD*10)
			{			
			avel.z = avel.z+drot;
			if (avel.z > RAD*10) avel.z = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD4))
		{
		if (avel.z > -RAD*10)
			{
			avel.z = avel.z-drot;
			if (avel.z < -RAD*10) avel.z = -RAD*10;
			}		
		}

		SetAngularVel(avel);
	}	
}

return 0;
}


void og2::clbkPreStep(double SimT, double SimDT, double MJD)
{

step_time = SimDT;///////
double simdt = SimDT;


//sprintf(oapiDebugString(), "%d %0.4f %0.4f", (sa1_stbd_elb_sts + sa1_stbd_scale_sts + sa1_port_elb_sts + sa1_port_scale_sts),sa1_port_rot_proc,sa1_stbd_rot_proc);

if(kr_status == 1)
{
VECTOR3 avel = _V(0,0,0);
GetAngularVel(avel);
double drot = RAD*2*simdt;

if (avel.x > -drot && avel.x < drot) avel.x = 0;
else if (avel.x > drot) avel.x = avel.x - drot;
else if (avel.x < -drot) avel.x = avel.x + drot;

if (avel.y > -drot && avel.y < drot) avel.y = 0;
else if (avel.y > drot) avel.y = avel.y - drot;
else if (avel.y < -drot) avel.y = avel.y + drot;

if (avel.z > -drot && avel.z < drot) avel.z = 0;
else if (avel.z > drot) avel.z = avel.z - drot;
else if (avel.z < -drot) avel.z = avel.z + drot;

if ((avel.x + avel.y + avel.z) == 0)kr_status = 0;

SetAngularVel(avel);
}



////////////////// ANIMATIONS ///////////////

// solp
if (solp_status == 1)
{
double da = SimDT * SOLP_SPEED;

solp_proc = solp_proc+da;
if (solp_proc > 1)
	{
	solp_proc = 1;
	solp_status = 2;
	}
SetAnimation(anim_solp, solp_proc);
}

else if (solp_status == 2)
{
SetAnimation(anim_solp, 1);
}



// masts
if (mast_status == 1)
{
double da = SimDT * MAST_SPEED;

mast_proc = mast_proc-da;
if (mast_proc < 0)
	{
	mast_proc = 0;
	mast_status = 0;
	}
SetAnimation(anim_mast, mast_proc);	
}

else if (mast_status == 2)SetAnimation(anim_mast, 1);	


////////////// sun tracking
if (solp_status == 2) // if panels are fully open
	{
	int a = 0;
	double Solpangle;
	VECTOR3 sunposglob = _V(0,0,0);
	VECTOR3 sunposloc = _V(0,0,0);
	Global2Local(sunposglob,sunposloc); // convert to vessel frame
	Solpangle = (PI*0.5*st_proc); // current rotation angle of panels
	normalise(sunposloc);

	double sunangle = acos(dotp(_V(0,1,0), sunposloc)); // current angle of sun  relative to y axis
	if (sunangle < 0) sunangle = -sunangle; // sanity check - angle always positive
	if (sunangle > (0.5*PI)) sunangle = (0.5*PI);

	double angle_diff = sunangle-Solpangle; // angular difference between sun and panels

	if (angle_diff > 0)a= 1;			// check which way the panels should rotate
	else if (angle_diff < 0)a=-1;	// check which way the panels should rotate
	else a = 0;

	double max_rot = SimDT * ST_SPEED * 0.5 * PI;	// check if angular difference
	if (angle_diff > 0 && angle_diff < max_rot) a = 0;			// is less than panel rotation
	else if (angle_diff < 0 && -angle_diff < max_rot) a = 0;	// during this frame

	// if a = 1 rotate one way
	// if a  = -1 rotate opposite way
	// if a  = 0 set panels exactly perpendicular to sun

	
		double db = SimDT * ST_SPEED;
		if (st_proc > 1) st_proc = (st_proc - 1); // range always 0-1
		else if (st_proc < 0) st_proc = (st_proc + 1); // range always 0-1
		else if (a == 1) st_proc = (st_proc+db);
		else if (a == -1) st_proc = (st_proc-db);
		else if (a == 0) st_proc = sunangle/(0.5*PI); // set exactly perpendicular to sun
		SetAnimation (anim_st, st_proc); // set the animation state and we're done!
	}


	if (asp_status == 1) DoSunPointing();
}

void og2::clbkPostStep (double simt, double simdt, double mjd)
{
	
}




bool og2::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);

	//int chw = 16;

	
	char test[256];
	strcpy(test,"abcdefghijklmnopqrstuvwxyz");
	int it = skp->GetTextWidth(test, 26);
	it = (int)(it/26);
	int chw = it;

	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);
	int w12 = (int)(376*d);
	int w13 = (int)(392*d);
	int w14 = (int)(408*d);
	int w15 = (int)(424*d);



	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(hProp);
	double test = ((m1+m2)/m1);
	double dv = (ISP*(log(test)));

	sprintf(abuf,"Remaining dV = %0.2f m/s", dv);
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"Reaction Wheel Att.Ctrl. = SHIFT+NUMPAD");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	if (kr_status == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"RW KillRot Active");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	if (asp_status == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Auto Attitude Active");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	if (asp_status == 0 && kr_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"B = Auto Attitude");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}



	if (solp_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"G = Deploy Solar Panels + Antenna");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}


	return true;
}


void og2::DoSunPointing()
{
	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);
	double dt = oapiGetSimStep();
	double max_acc = 2*RAD;
	double drot = RAD*2*dt;

	char cbuf[256]="Sun";
	OBJHANDLE hRef = oapiGetGbodyByName (cbuf);
	VECTOR3 sun_pos_global = _V(0,0,0);
	oapiGetGlobalPos(hRef, &sun_pos_global);

	char abuf[256]="Earth";
	OBJHANDLE eRef = oapiGetGbodyByName (abuf);
	VECTOR3 earth_pos_global = _V(0,0,0);
	oapiGetGlobalPos(eRef, &earth_pos_global);


	VECTOR3 sun_dir_local = _V(0,0,0);
	Global2Local(sun_pos_global,sun_dir_local);
	normalise(sun_dir_local);
	sun_dir_local = -sun_dir_local;

	VECTOR3 earth_dir_local = _V(0,0,0);
	Global2Local(earth_pos_global,earth_dir_local);
	normalise(earth_dir_local);

	VECTOR3 norm_ess_local = _V(0,0,0);
	norm_ess_local = crossp(sun_dir_local, earth_dir_local);
	normalise(norm_ess_local);

	VECTOR3 zfwd_local = _V(0,0,0);
	zfwd_local = crossp(earth_dir_local, norm_ess_local);
	normalise(zfwd_local);

	double xaxis_angle = atan2(earth_dir_local.x,-earth_dir_local.y);
	if (xaxis_angle < 0)xaxis_angle = xaxis_angle + (2*PI);

	double zaxis_angle = atan2(sun_dir_local.x,sun_dir_local.z);
	if (zaxis_angle < 0)zaxis_angle = zaxis_angle + (2*PI);

	double yaxis_angle = atan2(zfwd_local.y, zfwd_local.z);
	if (yaxis_angle < 0)yaxis_angle = yaxis_angle + (2*PI);

	double max_dec_ang = 0;
	double max_rot = 2*RAD;


	if (xaxis_angle < PI)
	{
	max_dec_ang = (avel.z/max_acc)*avel.z*0.5;
	if (max_dec_ang > xaxis_angle)
	{	
		if (avel.z < 0)avel.z = avel.z+drot;
		else avel.z = avel.z-drot;
	}
	else if (avel.z > -max_rot)avel.z = avel.z-drot;
	else avel.z = avel.z+drot;
		
	}
	else
	{
	max_dec_ang = (avel.z/max_acc)*avel.z*0.5;
	if (max_dec_ang > ((2*PI)-xaxis_angle))
		{	
		if (avel.z > 0)avel.z = avel.z-drot;
		else avel.z = avel.z+drot;
		}
	else if (avel.z < max_rot)avel.z = avel.z+drot;
	else avel.z = avel.z-drot;
	}



	if (yaxis_angle < PI)
	{
	max_dec_ang = (avel.x/max_acc)*avel.x*0.5;
	if (max_dec_ang > yaxis_angle)
		{	
		if (avel.x < 0)avel.x = avel.x+drot;
		else avel.x = avel.x-drot;
		}
	else if (avel.x < max_rot)avel.x = avel.x+drot;
	else avel.x = avel.x-drot;
	}
	else
	{
	max_dec_ang = (avel.x/max_acc)*avel.x*0.5;
	if (max_dec_ang > ((2*PI)-yaxis_angle))
		{	
		if (avel.x > 0)avel.x = avel.x-drot;
		else avel.x = avel.x+drot;
		}
	else if (avel.x > -max_rot)avel.x = avel.x-drot;
	else avel.x = avel.x+drot;
	}



	if (zaxis_angle < PI)
	{
	max_dec_ang = (avel.y/max_acc)*avel.y*0.5;
	if (max_dec_ang > zaxis_angle)
		{	
		if (avel.y < 0)avel.y = avel.y+drot;
		else avel.y = avel.y-drot;
		}
	else if (avel.y > -max_rot)avel.y = avel.y-drot;
	else avel.y = avel.y+drot;
	}
	else
	{
	max_dec_ang = (avel.y/max_acc)*avel.y*0.5;
	if (max_dec_ang > ((2*PI)-zaxis_angle))
		{	
		if (avel.y > 0)avel.y = avel.y-drot;
		else avel.y = avel.y+drot;
		}
	else if (avel.y < max_rot)avel.y = avel.y+drot;
	else avel.y = avel.y-drot;
	}

	SetAngularVel(avel);
	double ang_dif_x = xaxis_angle;
	if (ang_dif_x > PI)ang_dif_x = (2*PI)-xaxis_angle;

	double ang_dif_y = yaxis_angle;
	if (ang_dif_y > PI)ang_dif_y = (2*PI)-yaxis_angle;

	double ang_dif_z = zaxis_angle;
	if (ang_dif_z > PI)ang_dif_z = (2*PI)-zaxis_angle;

	
	if ((drot*5) > ang_dif_x && (drot*5) > ang_dif_y && (drot*5) > ang_dif_z)
	{


	VECTOR3 yup = _V(0,0,0);
	VECTOR3 zforward = _V(0,0,0);
	VECTOR3 xright = _V(0,0,0);

	GlobalRot(norm_ess_local, xright);
	GlobalRot(-earth_dir_local, yup);
	GlobalRot(zfwd_local,zforward);
	
	normalise(zforward);
	normalise(yup);
	normalise(xright);
	
	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);
	SetGlobalOrientation (_V(alpha, beta, (PI/2)-gamma));
	SetAngularVel (_V(0,0,0));

	//sprintf(oapiDebugString(), "yup %0.4f %0.4f %0.4f", yup.x, yup.y, yup.z);
	//sprintf(oapiDebugString(), "g %0.4f %0.4f", xright.x, yup.x);
	//sprintf(oapiDebugString(), "x y %0.3f %0.3f %0.3f %0.3f %0.3f %0.3f %0.3f %0.3f %0.3f", xright.x, xright.y, xright.z, yup.x, yup.y, yup.z, zforward.x, zforward.y, zforward.z);
	}
	//sprintf(oapiDebugString(), "yup %0.4f %0.4f %0.4f", yup.x, yup.y, yup.z);
}