// ==============================================================
//
// og2_espa2.cpp
// Control module for og2_espa2 vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>

// ==============================================================
// Some vessel parameters
// ==============================================================


const double DRYMASS = 200;

#define GREEN RGB(125, 255, 125)

static const DWORD ntdvtx_geardown = 3;
static TOUCHDOWNVTX tdvtx_geardown[ntdvtx_geardown] = {
	{_V( 0,-1, 5), 1e6, 1e6, 0.7, 0.7},
	{_V(-2,-1,-5), 1e6, 1e6, 0.7, 0.7},
	{_V( 2,-1,-5), 1e6, 1e6, 0.7, 0.7}
};

void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = CM[i] + (CM[i+1]-CM[i]) * f;  // aoa-dependent moment coefficient
	double saoa = sin(aoa);
	double pd = 0.03 + 0.4*saoa*saoa;  // profile drag
	*cd = pd + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	// profile drag + (lift-)induced drag + transonic/supersonic wave (compressibility) drag
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.03 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}
// ==============================================================
// Shuttle-PB class interface
// ==============================================================

class og2_espa2: public VESSEL3 {
public:
	og2_espa2 (OBJHANDLE hVessel, int flightmodel);
	~og2_espa2 ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void Jettison(void);


private:

	int npl_status;
	ATTACHMENTHANDLE og_attach[12];
	char nextpayload[256];
	
};

og2_espa2::og2_espa2 (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

og2_espa2::~og2_espa2 ()
{

}



// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void og2_espa2::clbkSetClassCaps (FILEHANDLE cfg)
{
	npl_status = 0;

	// physical vessel parameters
	SetSize (2);
	SetEmptyMass (DRYMASS);
	SetPMI(_V(2,2,0.75));
	SetCrossSections (_V(5,5,3));
	SetCW(1,1,1,1);
	SetRotDrag (_V(2,2,1));
	//SetTouchdownPoints ( _V(0,-1, 5), _V(-2,-1,-5), _V( 2,-1,-5));
	SetTouchdownPoints(tdvtx_geardown,ntdvtx_geardown);

	// camera parameters
	SetCameraOffset (_V(0,2,-2));

	SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/og2_espa2"),MESHVIS_ALWAYS);


	// propellant resources
	

	// ***************** thruster definitions *******************
	

	///////////////////////////////////////////////////////
	/////AirFoils

	CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,1), VLiftCoeff, 0, 1, 2, 1);
	CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,1), HLiftCoeff, 0, 1, 2, 1);

	 og_attach[0] = CreateAttachment (false, _V( 0.987, 0, 1.09), _V( 1,0,0), _V(0,0,1), "PL1");
	 og_attach[1] = CreateAttachment (false, _V( 0, 0.987, 1.09), _V(0, 1,0), _V(0,0,1), "PL2");
	 og_attach[2] = CreateAttachment (false, _V(-0.987, 0, 1.09), _V(-1,0,0), _V(0,0,1), "PL3");
	 og_attach[3] = CreateAttachment (false, _V( 0,-0.987, 1.09), _V(0,-1,0), _V(0,0,1), "PL4");

	 og_attach[4] = CreateAttachment (false, _V( 0.987, 0, 0), _V( 1,0,0), _V(0,0,1), "PL5");
	 og_attach[5] = CreateAttachment (false, _V( 0, 0.987, 0), _V(0, 1,0), _V(0,0,1), "PL6");
	 og_attach[6] = CreateAttachment (false, _V(-0.987, 0, 0), _V(-1,0,0), _V(0,0,1), "PL7");
	 og_attach[7] = CreateAttachment (false, _V( 0,-0.987, 0), _V(0,-1,0), _V(0,0,1), "PL8");

	 og_attach[8] = CreateAttachment (false, _V( 0.987, 0, -1.09), _V( 1,0,0), _V(0,0,1), "PL9");
	 og_attach[9] = CreateAttachment (false, _V( 0, 0.987, -1.09), _V(0, 1,0), _V(0,0,1), "PL10");
	 og_attach[10] = CreateAttachment (false, _V(-0.987, 0, -1.09), _V(-1,0,0), _V(0,0,1), "PL11");
	 og_attach[11] = CreateAttachment (false, _V( 0,-0.987, -1.09), _V(0,-1,0), _V(0,0,1), "PL12");

}



// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new og2_espa2 (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (og2_espa2*)vessel;
}


void og2_espa2::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState(scn);

	//sprintf (cbuf, "%d %0.4f %d %0.4f", solp_status, solp_proc, st_status, st_proc);
	//oapiWriteScenario_string (scn, "SOLP", cbuf);

	//sprintf (cbuf, "%d %0.4f", mast_status, mast_proc);
	//oapiWriteScenario_string (scn, "LDMA", cbuf);
	
	sprintf (cbuf, "%d", npl_status);
	oapiWriteScenario_string (scn, "NPLS", cbuf);


}
void og2_espa2::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "NPLS", 4)) {
			sscanf (line+4, "%d", &npl_status);
		//} else if (!strnicmp (line, "LDMA", 4)) {
		//	sscanf (line+4, "%d%lf", &mast_status, &mast_proc);
		//} else if (!strnicmp (line, "KROT", 4)) {
		//	sscanf (line+4, "%d%d",&kr_status, &asp_status);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }




}



int og2_espa2::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {



if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {	 
	switch (key)
	{
		

	} 
  }

  else if (KEYMOD_CONTROL (kstate)) 
  { 
	switch (key)
	{
		
	} 
  }

else { // unmodified keys
    switch (key) {

		case OAPI_KEY_N: // cycle next payload
		if (npl_status == 11) npl_status = 0;
		else npl_status = npl_status + 1;
		return 1;

		case OAPI_KEY_J: // jettison
		Jettison();
		return 1;


	}
  }
  return 0;
}



void og2_espa2::clbkPreStep(double SimT, double SimDT, double MJD)
{
	double pld_mass =0.0;

	for (DWORD j = 0; j < 12; j++)
	{
	OBJHANDLE hOg = GetAttachmentStatus(og_attach[j]);
	if (hOg == NULL) continue;
	else pld_mass = pld_mass + oapiGetMass(hOg);
	}

	double total_mass = DRYMASS + pld_mass;
	SetEmptyMass(total_mass);

//sprintf(oapiDebugString(), "%d %0.4f %0.4f", (sa1_stbd_elb_sts + sa1_stbd_scale_sts + sa1_port_elb_sts + sa1_port_scale_sts),sa1_port_rot_proc,sa1_stbd_rot_proc);
}

void og2_espa2::clbkPostStep (double simt, double simdt, double mjd)
{
	
}

void og2_espa2::Jettison(void)
{
DetachChild(og_attach[npl_status], 0.3);
npl_status = npl_status+1;
if (npl_status == 12)npl_status = 0;

}


bool og2_espa2::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);

	//int chw = 16;

	
	char test[256];
	strcpy(test,"abcdefghijklmnopqrstuvwxyz");
	int it = skp->GetTextWidth(test, 26);
	it = (int)(it/26);
	int chw = it;

	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	//int w2 = (int)(216*d);
	//int w3 = (int)(232*d);
	//int w4 = (int)(248*d);
	//int w5 = (int)(264*d);
	//int w6 = (int)(280*d);
	//int w7 = (int)(296*d);
	//int w8 = (int)(312*d);
	//int w9 = (int)(328*d);
	//int w10 = (int)(344*d);
	//int w11 = (int)(360*d);
	//int w12 = (int)(376*d);
	//int w13 = (int)(392*d);
	//int w14 = (int)(408*d);
	int w15 = (int)(424*d);



	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"J = Jettison Payload");
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"N = Select Next Payload");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	OBJHANDLE hOg = GetAttachmentStatus(og_attach[npl_status]);
	if (hOg == NULL) strcpy(abuf, "none");
	else oapiGetObjectName(hOg, abuf, 60); 
	char bbuf[256];
	strcpy(bbuf, "Next Payload = ");
	strcat(bbuf, abuf);
	skp->Text ((10+roxl), (w1+royl), bbuf,strlen(bbuf));
	}

	
	return true;
}
