// ==============================================================
//                 ORBITER MODULE: slc39ar
//
// slc39ar.cpp
// Control module for slc39ar vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "gcAPI.h"

// ==============================================================
// Some vessel parameters
// ==============================================================


const double CLAWS_SPEED = 0.05;
const double SLOW_RETRACT_SPEED = 0.002;
const double FAST_RETRACT_SPEED = 0.50;
const double PB_FUELMASS = 750.0;
const double PB_ISP = 5e4;
const double PB_MAXHOVERTH = 0;



static const DWORD ntdvtx_geardown = 3;
static TOUCHDOWNVTX tdvtx_geardown[ntdvtx_geardown] = {
	{_V(0,-21.5,100), 1e6, 1e6, 0.7, 0.7},
	{_V(-86.6,-21.5,-50), 1e6, 1e6, 0.7, 0.7},
	{_V(86.6,-21.5,-50), 1e6, 1e6, 0.7, 0.7}
};


class slc39ar: public VESSEL3 {
public:
	slc39ar (OBJHANDLE hVessel, int flightmodel)
		: VESSEL3 (hVessel, flightmodel) {}
	void DefineAnimations (void);
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);

	THRUSTER_HANDLE th_hover, th_steam[2];
	ANIMATIONCOMPONENT_HANDLE Section0, Section1;
	ATTACHMENTHANDLE fh_attach;


private:
	UINT anim_body_rot, anim_pist_rot, anim_pist_trans, anim_claws;
	UINT anim_umb_rot;
	double body_rot_proc, pist_rot_proc, pist_trans_proc, claws_proc;
	double umb_rot_proc;
	int light_status, lox_status, launch_status, stbk_status;
	SpotLight *slca1_light;
	SpotLight *slca2_light;
};
// Define animation sequences for moving parts
// --------------------------------------------------------------
void slc39ar::DefineAnimations (void)
{

	//  TOWER 
	anim_body_rot = CreateAnimation (0);
	static UINT ChuteGrp1[4] = {0,1,9,10};

	static MGROUP_ROTATE  ChuteFold1 (0, ChuteGrp1, 4, _V(0,-2.495,-4.755),_V(1,0,0), -30*RAD);
	Section0 = AddAnimationComponent (anim_body_rot, 0, 1,   &ChuteFold1);

	// claws
	anim_claws = CreateAnimation (0);
	static UINT ChuteGrp2[1] = {8};

	static MGROUP_ROTATE  ChuteFold2 (0, ChuteGrp2, 1, _V(-1.036,0,-2.423),_V(0,1,0), -50*RAD);
	AddAnimationComponent (anim_claws, 0, 1,   &ChuteFold2, Section0);

	static UINT ChuteGrp3[1] = {7};

	static MGROUP_ROTATE  ChuteFold3 (0, ChuteGrp3, 1, _V(1.036,0,-2.423),_V(0,1,0), 50*RAD);
	AddAnimationComponent (anim_claws, 0, 1,   &ChuteFold3, Section0);


	// pistons
	anim_pist_rot = CreateAnimation (0);
	static UINT ChuteGrp4[1] = {2};

	static MGROUP_ROTATE  ChuteFold4 (0, ChuteGrp4, 1, _V(0,-13.285,-6.901),_V(1,0,0), -90*RAD);
	Section1 = AddAnimationComponent (anim_pist_rot, 0, 1,   &ChuteFold4);

	anim_pist_trans = CreateAnimation (0);
	static UINT ChuteGrp5[1] = {3};

	VECTOR3 pist_trans = _V(0,-18.829,-0.024); // length
	normalise(pist_trans);
	pist_trans.x = 10*pist_trans.x;
	pist_trans.y = 10*pist_trans.y;
	pist_trans.z = 10*pist_trans.z; // 10m trans

	static MGROUP_TRANSLATE  ChuteFold5 (0, ChuteGrp5, 1, pist_trans);
	AddAnimationComponent (anim_pist_trans, 0, 1,   &ChuteFold5, Section1);


	///// umb
	
	static UINT UmbGrp1[1] = {4};
	anim_umb_rot = CreateAnimation (0);

	static MGROUP_ROTATE  Umb1 (0, UmbGrp1, 1, _V(0,-2.495,-4.755),_V(1,0,0), 30*RAD);
	AddAnimationComponent (anim_umb_rot, 0, 1,   &Umb1, Section0);

}


// ==============================================================
// slc39ar class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void slc39ar::clbkSetClassCaps (FILEHANDLE cfg)
{
	body_rot_proc = 0.0;
	pist_rot_proc = 0.0;
	pist_trans_proc = 0.0;
	claws_proc = 0.0;
	umb_rot_proc = 0.0;

	stbk_status = 0;
	light_status = 0;
	lox_status = 0;
	launch_status = 0;

	DefineAnimations ();
	gcInitialize();
	// physical specs
	SetEmptyMass (1000000);
	SetCW (1, 1, 1, 1);
	SetSize (200);
	SetCrossSections (_V(1000,1000,10000));
	SetPMI (_V(100,100,100));
	//SetTouchdownPoints (_V(0,-21.5,100), _V(-86.6,-21.5,-50), _V(86.6,-21.5,-50));
	SetTouchdownPoints (tdvtx_geardown,ntdvtx_geardown);
	SetRotDrag (_V(1,1,1));

	SetMeshVisibilityMode ((AddMesh ("Falcon9_BLOCK4/lc39a_strongback_r")), MESHVIS_ALWAYS);
	SetCameraOffset (_V(0,10,-20));

	PROPELLANT_HANDLE ph_vernier = CreatePropellantResource (PB_FUELMASS);

	PARTICLESTREAMSPEC contrail_hover = {
		0, 0.2, 15, 0.5, 10, 1.5, 1, 0.5, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC contrail_hover2 = {
		0, 1, 15, 1.5, 10, 2.5, 1.5, 3, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	// big launch smoke
	SURFHANDLE boom1 = oapiRegisterParticleTexture ("Contrail4");

	PARTICLESTREAMSPEC contrail_steam = {
		0, 40, 10, 250, 1.3, 1.5, 16, 3, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC contrail_steam3 = {
		0, 30, 6, 250, 0.8, 0.8, 20, 3, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 4,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1,
		boom1
	};

	if (gcEnabled() == true)
	{
		contrail_steam.atmslowdown = 2.5;
		contrail_steam.lifetime = 2;
		contrail_steam3.atmslowdown = 2.5;
		contrail_steam3.lifetime = 1.1;
	}

// small launch smoke
	PARTICLESTREAMSPEC contrail_steam2 = {
		0, 10, 30, 100, 0.75, 0.5, 6, 3, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	th_hover = CreateThruster (_V(0,0,0), _V(0,0,1), PB_MAXHOVERTH, ph_vernier, PB_ISP);

	//th_steam = CreateThruster (_V(0,0,0), _V(0,0,-1), PB_MAXHOVERTH, ph_vernier, PB_ISP);

	th_steam[0] = CreateThruster (_V(0, 0, 50), _V(0,-0.259,-0.966), PB_MAXHOVERTH, ph_vernier, PB_ISP);
	AddExhaustStream (th_steam[0], &contrail_steam);
	AddExhaustStream (th_steam[0], &contrail_steam3);

	th_steam[1] = CreateThruster (_V(0, -2, 0), _V(0, -1, 0), PB_MAXHOVERTH, ph_vernier, PB_ISP);
	AddExhaustStream (th_steam[1], &contrail_steam2);

	
	if (gcEnabled() == true) AddExhaustStream (th_hover, _V(0,41,-1.8), &contrail_hover2);
	else AddExhaustStream (th_hover, _V(0,41,-1.8), &contrail_hover);
	//AddExhaustStream (th_hover, _V(0,52.2,-1.8), &contrail_hover);

	//AddExhaustStream (th_steam, _V(0,8,0), &contrail_steam);


	// ***************** thruster definitions *******************

	COLOUR4 col_d = {1,1,0.9,0};
	COLOUR4 col_s = {1,1,0.9,0};
	COLOUR4 col_a = {0,0,0,0};
	COLOUR4 col_white = {1,1,1,0};
	slca1_light = (SpotLight*)AddSpotLight(_V(25,10,70), _V(-0.612,0.5,-0.612), 4000, 2e-4, 0, 2e-4, RAD*90, RAD*180, col_d, col_s, col_a);
	slca1_light->Activate(false);

	slca2_light = (SpotLight*)AddSpotLight(_V(-25,10,-80), _V(0.612,0.5,0.612), 4000, 2e-4, 0, 2e-4, RAD*90, RAD*180, col_d, col_s, col_a);
	slca2_light->Activate(false);

	

	fh_attach = CreateAttachment (false, _V( 0, 0, 0), _V( 0,1,0), _V(0,0,-1), "FH");


}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new slc39ar (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (slc39ar*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void slc39ar::clbkPostCreation ()
{
	
}


void slc39ar::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	sprintf (cbuf, "%0.4f %0.4f %0.4f %0.4f", body_rot_proc, pist_rot_proc, pist_trans_proc, claws_proc);
	oapiWriteScenario_string (scn, "ANM", cbuf);
	sprintf (cbuf, "%d %d", stbk_status, launch_status);
	oapiWriteScenario_string (scn, "STTS", cbuf);
}
void slc39ar::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "ANM", 3)) {
			sscanf (line+3, "%lf%lf%lf%lf", &body_rot_proc, &pist_rot_proc, &pist_trans_proc, &claws_proc);
		} else if (!strnicmp (line, "STTS", 4)) {
			sscanf (line+4, "%d%d", &stbk_status, &launch_status);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_claws, claws_proc);
	SetAnimation (anim_body_rot, body_rot_proc);
	
}


int slc39ar::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {
  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) {

  } else { // unmodified keys
    switch (key) {

	case OAPI_KEY_G: // move mss
		if(stbk_status == 0)stbk_status = 1;
		else if(stbk_status == 3)stbk_status = 6;
		else if(stbk_status == 5)stbk_status = 6;
		return 1;

	case OAPI_KEY_K: // lights on/off
		if (light_status == 0) light_status = 1;
		else if (light_status == 1) light_status = 0;
		return 1;

	case OAPI_KEY_V: // LOX vent on/off
		if (lox_status == 0)lox_status = 1;
		else lox_status = 0;
		return 1;
	}
  }
  return 0;
}
void slc39ar::clbkPreStep(double SimT, double SimDT, double MJD) {

	if (stbk_status == 1)
	{
		claws_proc = claws_proc+(CLAWS_SPEED*SimDT);
		if (claws_proc > 1)
		{
			claws_proc = 1;
			stbk_status = 2;
		}
	}
	if (stbk_status == 2)
	{
		body_rot_proc = body_rot_proc+(SLOW_RETRACT_SPEED*SimDT);
		if (body_rot_proc > 0.05)
		{
			body_rot_proc = 0.05;
			stbk_status = 3;
		}
		umb_rot_proc = body_rot_proc;
	}
	if (stbk_status == 3)
	{
		umb_rot_proc = body_rot_proc;
		if (launch_status == 3)stbk_status = 4;
	}
	if (stbk_status == 4)
	{
		body_rot_proc = body_rot_proc+(FAST_RETRACT_SPEED*SimDT);
		if (body_rot_proc > 1)
		{
			body_rot_proc = 1;
			stbk_status = 5;
		}
		umb_rot_proc = umb_rot_proc-(0.1 * SimDT);
		if(umb_rot_proc < 0)umb_rot_proc = 0;
	}
	//
	if (stbk_status == 6)
	{
		body_rot_proc = body_rot_proc-(SLOW_RETRACT_SPEED*SimDT);
		if (umb_rot_proc > body_rot_proc) umb_rot_proc = body_rot_proc;
		if(body_rot_proc < 0)
		{
			body_rot_proc = 0;
			umb_rot_proc = 0;
			stbk_status = 7;
		}
	}
	if (stbk_status == 7)
	{
		claws_proc = claws_proc-(CLAWS_SPEED*SimDT);
		if (claws_proc < 0)
		{
			claws_proc = 0;
			stbk_status = 0;
		}
	}
	SetAnimation (anim_claws, claws_proc);
	SetAnimation (anim_body_rot, body_rot_proc);
	SetAnimation (anim_umb_rot, umb_rot_proc);

	//// pistons
	VECTOR3 pist_top = _V(0,5.544,-6.877);
	VECTOR3 pist_btm = _V(0,-13.285,-6.901);
	VECTOR3 orig_pist_vector = (pist_top - pist_btm);
	double orig_pist_length = length(orig_pist_vector);
	VECTOR3 main_pivot = _V(0,-2.495,-4.755);
	VECTOR3 piv_to_pist = pist_top - main_pivot;

	double theta = body_rot_proc*30*RAD;
	double sin_theta = sin(theta);
	double cos_theta = cos(theta);

	VECTOR3 pist_top_new = _V(0,0,0);

	pist_top_new.z = (piv_to_pist.z*cos_theta) - (piv_to_pist.y*sin_theta);
	pist_top_new.y = (piv_to_pist.z*sin_theta) + (piv_to_pist.y*cos_theta);
	pist_top_new = pist_top_new + main_pivot;

	VECTOR3 new_pist_vector = (pist_top_new - pist_btm);
	double new_pist_length = length(new_pist_vector);

	double req_trans = orig_pist_length - new_pist_length;
	normalise(new_pist_vector);
	normalise(orig_pist_vector);
	double req_rot = acos(dotp(new_pist_vector,orig_pist_vector));

	pist_trans_proc = req_trans*0.1;
	pist_rot_proc = req_rot/(0.5*PI);

	SetAnimation (anim_pist_rot, pist_rot_proc);
	SetAnimation (anim_pist_trans, pist_trans_proc);



	// ///////////////////////////////////
	ENGINESTATUS es;
	VECTOR3 gpos = _V(0,0,0);
	VECTOR3 gpad = _V(0,0,0);
	Local2Global(_V(0,40,0),gpad);

	es.main = 0;
	
	if (lox_status == 0)
{
	for (DWORD i = 0; i < oapiGetVesselCount(); i++)
	{
	OBJHANDLE hV = oapiGetVesselByIndex (i);

	if (hV == GetHandle()) continue; // we don't want to count ourselves ...
	oapiGetGlobalPos (hV, &gpos);

	if (dist (gpos, gpad) < 50)
	{ // in range
		SetThrusterLevel (th_hover, 1);
		VESSEL *v = oapiGetVesselInterface (hV);
		oapiGetEngineStatus(hV, &es);
		if (es.main > 0) 
		{
			SetThrusterLevel (th_hover, 0);
			break;
		}
	}

	}
}
	else SetThrusterLevel (th_hover, 0);
	if(oapiGetTimeAcceleration() > 10)SetThrusterLevel (th_hover, 0);
	
	if (es.main > 0 ) 
	{
		SetThrusterLevel (th_steam[0], 1);
		SetThrusterLevel (th_steam[1], 1);
		launch_status = 1;
		if (es.main > 0.9 )launch_status = 2;
	}
	else 
	{
		SetThrusterLevel (th_steam[0], 0);
		SetThrusterLevel (th_steam[1], 0);
	}

	if(oapiGetTimeAcceleration() > 10)
	{
		SetThrusterLevel (th_hover, 0);
		SetThrusterLevel (th_steam[1], 0);
		SetThrusterLevel (th_steam[1], 0);
	}
	if (es.main > 0.95 )
	{
		DetachChild(fh_attach, 0.0);
		launch_status = 3;
	}

/////////////////////shadowcheck
	VECTOR3 rsuninv;
	VECTOR3 rsunloc;
	VECTOR3 rsun;
	GetGlobalPos(rsun);
	rsuninv = _V(-rsun.x, -rsun.y, -rsun.z);
	Global2Local(rsuninv,rsunloc);

	if (slca1_light->IsActive()== false && rsunloc.y < 0)
			{
			slca1_light->Activate(true);
			slca2_light->Activate(true);
			}
	else if (slca1_light->IsActive()== true && rsunloc.y > 0)
			{
			slca1_light->Activate(false);
			slca2_light->Activate(false);
			}

	if (light_status == 1)
			{
			slca1_light->Activate(false);
			slca2_light->Activate(false);
			}
	if (launch_status == 1)
			{
			slca1_light->SetIntensity(1-(es.main));
			slca2_light->SetIntensity(1-(es.main));
			}
	else if (launch_status > 1)
			{
			slca1_light->Activate(false);
			slca2_light->Activate(false);
			}
	}

void slc39ar::clbkPostStep (double simt, double simdt, double mjd)
{
	
	
}